import argparse
import traceback
import sys
from pathlib import Path

from _pyproject_hooks import BuildBackend, BackendHookError


def parse_config_settings_args(config_settings):
    """
    Given a list of config `KEY=VALUE` formatted config settings,
    return a dictionary that can be passed to PEP 517 hook functions.
    """
    if not config_settings:
        return config_settings
    new_config_settings = {}
    for arg in config_settings:
        key, _, value = arg.partition('=')
        if key in new_config_settings:
            if not isinstance(new_config_settings[key], list):
                # convert the existing value to a list
                new_config_settings[key] = [new_config_settings[key]]
            new_config_settings[key].append(value)
        else:
            new_config_settings[key] = value
    return new_config_settings


def build_wheel(*, wheeldir, config_settings=None):
    wheeldir.mkdir(parents=True, exist_ok=True)
    BuildBackend.from_directory().get_caller().build_wheel(wheeldir, config_settings)


def parse_args(argv=None):
    parser = argparse.ArgumentParser(prog='%pyproject_wheel')
    parser.add_argument('wheeldir', help=argparse.SUPPRESS, type=Path)
    parser.add_argument(
        '-C',
        dest='config_settings',
        action='append',
        help='Configuration settings to pass to the PEP 517 backend',
    )
    args = parser.parse_args(argv)
    args.config_settings = parse_config_settings_args(args.config_settings)
    return args


if __name__ == '__main__':
    try:
        build_wheel(**vars(parse_args()))
    except BackendHookError as exc:
        print(exc, file=sys.stderr)
        sys.exit(exc.returncode)
    except Exception:
        traceback.print_exc()
        sys.exit(1)
    else:
        sys.exit(0)
